#include <stdlib.h>
#include <stdio.h>

#include "wfm.h"

#define Ncomplex 2
#define Ncolour 3
#define Nspin2 2
#define Nspin4 4
#define Ncb 2
#define ND 4
#define Re 0
#define Im 1

#define   MYRAND_MAX        (2147483647)

unsigned long long rand_seed;
void mysrand(unsigned int seed)
{
  rand_seed = seed;
}
 
int myrand(void)
{
  int r;
 
  rand_seed = rand_seed *  6364136223846793005LL + 1;
  return (int) ( ( rand_seed  >> 32 ) & MYRAND_MAX );
}

void print_result(char *name,Float *ans, int nword);
void print_result(char *name,Float *ans, int nword)
{
  int i;
  FILE *result;
  result = fopen(name,"w");
  for (i=0;i<nword;i++) {
    fprintf(result,"%16.8le\n",ans[i]);
  }
  fclose(result);
}

struct Spinor { 
  Float vals[ND][Ncolour][Ncomplex]; 
};
typedef struct Spinor Spin4;


typedef struct { Float vals[Nspin4][Ncolour][Ncolour][Ncomplex];} Gauge;

void RanFloat(Float *,int n);
void RanFloat(Float *f,int n)
{
  for ( int i=0;i<n;i++ ) {
    f[i] = (Float)myrand()/MYRAND_MAX;
  }
}
void UnitMatrix(Float *,int n);
void UnitMatrix(Float *f,int n)
{
  Float *U = f;
  for ( int i=0;i<n;i++ ) {
    int mu = i%4;
    for ( int r=0;r<3;r++ ) {
    for ( int c=0;c<3;c++ ) {
      if ( (r == c) && (mu == 2) ) 
	*(f++) = 1.0;
      else
	*(f++) = 0.0;
      *(f++) = 0.0;
    }
    }
  }
  int r=0;
  int c=2;
  //U[2*c+r*6] = 10.0;
  //U[2*c+r*6+1] = 0.0;
}

int main (int argc,char **argv )
{
#ifdef USE_COMMS_QMP
  QMP_thread_level_t thread_level_required=QMP_THREAD_SINGLE;
  QMP_thread_level_t thread_level_provided;
  QMP_init_msg_passing(&argc, &argv, thread_level_required, &thread_level_provided);
#endif

  int lx,ly,lz,lt;
  int lxyzt;
  int sx;

#ifndef USE_COMMS_QMP
  if ( argc != 6 ) { 
    printf("Usage: %s lx ly lz lt bgl=0,1\n All must be even",argv[0]);
    exit(-1);
  }
  lx = atoi(argv[1]);
  ly = atoi(argv[2]);
  lz = atoi(argv[3]);
  lt = atoi(argv[4]);
  int bgl = atoi(argv[5]);
  wfm::bgl_kernels = bgl;
#else 
  if (argc != 9 ) { 
    printf("Usage: %s lx ly lz lt px py pz pt\n",argv[0]);
    exit(-1);
  }
  lx = atoi(argv[1]);
  ly = atoi(argv[2]);
  lz = atoi(argv[3]);
  lt = atoi(argv[4]);
  int dims[4];
  for(int i=0; i < 4; i++) { 
    dims[i] = atoi(argv[5+i]);
  }
  QMP_status_t ltopP=QMP_declare_logical_topology(dims, 4);
  if (ltopP != QMP_SUCCESS) { 
    fprintf(stderr, "Failed to declare logical topology mate \n");
    exit(-1);
  }
#endif

  if( (lx&0x1)||(ly&0x1)||(lz&0x1)||(lt&0x1) ) { 
    printf("Usage: %s lx ly lz lt\nAll must be even\n",argv[0]);
    printf("%d %d %d %d\n",lx,ly,lz,lt);
    fflush(stdout);
    exit(-1);
  }

  sx = lx/2;
  lxyzt = sx*ly*lz*lt;

  Spin4 *Chi = new Spin4[lxyzt*Ncb];
  Spin4 *PSI = new Spin4[lxyzt*Ncb];
  Gauge *U = new Gauge[lxyzt*Ncb];

  RanFloat((Float *)Chi,lxyzt*24*Ncb);
  RanFloat((Float *)PSI,lxyzt*24*Ncb);

  RanFloat((Float *)U,lxyzt*18*4*Ncb);
  //UnitMatrix((Float *)U,lxyzt*4*Ncb);

  WilsonArg wil;

  wil.local_latt[0]  = lx;
  wil.local_latt[1]  = ly;
  wil.local_latt[2]  = lz;
  wil.local_latt[3]  = lt;

#ifdef USE_COMMS_NONE
  wil.local_comm[0]  = 1;
  wil.local_comm[1]  = 1;
  wil.local_comm[2]  = 1;
  wil.local_comm[3]  = 1;
#else
  wil.local_comm[0]  = 0;
  wil.local_comm[1]  = 0;
  wil.local_comm[2]  = 0;
  wil.local_comm[3]  = 0;
#endif

  printf("local_comm[0] = %d\n", wil.local_comm[0]);
  printf("local_comm[1] = %d\n", wil.local_comm[1]);
  printf("local_comm[2] = %d\n", wil.local_comm[2]);
  printf("local_comm[3] = %d\n", wil.local_comm[3]);

  wfm_init(&wil);

  wfm_dslash((Float *)Chi, 
	     (Float *)U, 
	     (Float *)PSI, 
	     0,
	     0);

  print_result("Chi_cb0",(Float *)Chi,24*lxyzt);

 
  wfm_dslash((Float *)Chi, 
	     (Float *)U, 
	     (Float *)PSI, 
	     0,
	     1);

  
  print_result("Chi_cb0_dag",(Float *)Chi,24*lxyzt);
  wfm_dslash((Float *)Chi, 
		(Float *)U, 
		(Float *)PSI, 
		1,
		0);
  print_result("Chi_cb1",(Float *)Chi,24*lxyzt);
  wfm_dslash((Float *)Chi, 
		(Float *)U, 
		(Float *)PSI, 
		1,
		1);
  print_result("Chi_cb1_dag",(Float *)Chi,24*lxyzt);
  wfm_end(&wil);

#ifdef USE_COMMS_QMP
  QMP_finalize_msg_passing();
#endif
  exit(0);
}







